import json
import os
import uuid
import tempfile
import subprocess

from spark_etl.job_submitters import AbstractJobSubmitter

class PySparkJobSubmitter(AbstractJobSubmitter):
    def __init__(self, config):
        super(PySparkJobSubmitter, self).__init__(config)


    def run(self, deployment_location, options={}, args={}, handlers=[], on_job_submitted=None):
        # version is already baked into deployment_location
        # local submitter ignores handlers
        run_id  = str(uuid.uuid4())
        run_dir = self.config['run_dir']
        app_dir = deployment_location

        os.makedirs(os.path.join(run_dir, run_id))

        # generate input.json
        with open(os.path.join(run_dir, run_id, 'input.json'), 'wt') as f:
            json.dump(args, f)

        subprocess.check_call(["spark-submit", os.path.join(deployment_location, "job_loader.py"), "--run-id", run_id, "--run-dir", run_dir, "--app-dir", app_dir])
        if on_job_submitted is not None:
            on_job_submitted(run_id, vendor_info={})
            
        with open(os.path.join(run_dir, run_id, "result.json"), "r") as f:
            return json.load(f)



import json
from urllib.parse import urlparse
import time
import uuid
from datetime import datetime, timedelta
import os
from termcolor import colored, cprint
import readline

import oci
from oci_core import get_os_client, get_df_client, os_upload, os_upload_json, os_download, os_download_json, os_get_endpoint, \
    os_has_object, os_delete_object

from spark_etl.job_submitters import AbstractJobSubmitter
from spark_etl import SparkETLLaunchFailure, SparkETLGetStatusFailure, SparkETLKillFailure
from .tools import check_response, remote_execute
from spark_etl.utils import CLIHandler, handle_server_ask
from spark_etl.core import ClientChannelInterface

class ClientChannel(ClientChannelInterface):
    def __init__(self, region, oci_config, run_base_dir, run_id):
        self.region = region
        self.oci_config = oci_config
        self.run_base_dir = run_base_dir
        self.run_id = run_id

        o = urlparse(run_base_dir)
        self.namespace = o.netloc.split('@')[1]
        self.bucket = o.netloc.split('@')[0]
        self.root_path = o.path[1:] # remove the leading "/"


    def read_json(self, name):
        os_client = get_os_client(self.region, self.oci_config)
        object_name = os.path.join(self.root_path, self.run_id, name)
        result = os_download_json(os_client, self.namespace, self.bucket, object_name)
        return result


    def write_json(self, name, payload):
        os_client = get_os_client(self.region, self.oci_config)
        object_name = os.path.join(self.root_path, self.run_id, name)
        os_upload_json(os_client, payload, self.namespace, self.bucket, object_name)


    def has_json(self, name):
        os_client = get_os_client(self.region, self.oci_config)
        object_name = os.path.join(self.root_path, self.run_id, name)
        return os_has_object(os_client, self.namespace, self.bucket, object_name)


    def delete_json(self, name):
        os_client = get_os_client(self.region, self.oci_config)
        object_name = os.path.join(self.root_path, self.run_id, name)
        os_delete_object(os_client, self.namespace, self.bucket, object_name)


class DataflowJobSubmitter(AbstractJobSubmitter):
    def __init__(self, config):
        super(DataflowJobSubmitter, self).__init__(config)
        # config fields
        # region, e.g. IAD
        # run_base_dir, uri, point to the run directory.
        run_base_dir = self.config['run_base_dir']
        o = urlparse(run_base_dir)
        if o.scheme != 'oci':
            raise SparkETLLaunchFailure("run_base_dir must be in OCI")


    @property
    def region(self):
        return self.config['region']



    def run(self, deployment_location, options={}, args={}, handlers=None, on_job_submitted=None, cli_mode=False):
        # options fields
        #     num_executors     : number
        #     driver_shape      :  string
        #     executor_shape    : string
        #     lib_url_duration  : number (repre the number of minutes)
        #     on_job_submitted  : callback, on_job_submitted(run_id, vendor_info={'oci_run_id': 'xxxyyy'})

        o = urlparse(deployment_location)
        if o.scheme != 'oci':
            raise SparkETLLaunchFailure("deployment_location must be in OCI")

        run_base_dir = self.config['run_base_dir']
        run_id = str(uuid.uuid4())

        namespace = o.netloc.split('@')[1]
        bucket = o.netloc.split('@')[0]
        root_path = o.path[1:] # remove the leading "/"

        # let's get the deployment.json
        os_client = get_os_client(self.region, self.config.get("oci_config"))
        deployment = os_download_json(os_client, namespace, bucket, os.path.join(root_path, "deployment.json"))

        # let's upload the args
        client_channel = ClientChannel(
            self.region,
            self.config.get("oci_config"),
            run_base_dir,
            run_id
        )
        client_channel.write_json("args.json", args)

        o = urlparse(self.config['run_base_dir'])
        namespace = o.netloc.split('@')[1]
        bucket = o.netloc.split('@')[0]
        root_path = o.path[1:] # remove the leading "/"
        os_upload_json(os_client, args, namespace, bucket, f"{root_path}/{run_id}/args.json")

        df_client = get_df_client(self.region, self.config.get("oci_config"))
        crd_argv = {
            'compartment_id': deployment['compartment_id'],
            'application_id': deployment['application_id'],
            'display_name' :options["display_name"],
            'arguments': [
                "--deployment-location", deployment_location,
                "--run-id", run_id,
                "--run-dir", os.path.join(run_base_dir, run_id),
                "--app-region", self.region,
            ],
        }
        for key in ['num_executors', 'driver_shape', 'executor_shape']:
            if key in options:
                crd_argv[key] = options[key]

        create_run_details = oci.data_flow.models.CreateRunDetails(**crd_argv)
        r = df_client.create_run(create_run_details=create_run_details)
        check_response(r, lambda : SparkETLLaunchFailure("dataflow failed to run the application"))
        run = r.data
        oci_run_id = run.id
        print(f"Job launched, run_id = {run_id}, oci_run_id = {run.id}")
        if on_job_submitted is not None:
            on_job_submitted(run_id, vendor_info={'oci_run_id': run.id})

        cli_entered = False
        while True:
            time.sleep(10)
            r = df_client.get_run(run_id=run.id)
            check_response(r, lambda : SparkETLGetStatusFailure("dataflow failed to get run status"))
            run = r.data
            print(f"Status: {run.lifecycle_state}")
            if run.lifecycle_state in ('FAILED', 'SUCCEEDED', 'CANCELED'):
                break
            handle_server_ask(client_channel, handlers)

            if cli_mode and not cli_entered and run.lifecycle_state == 'IN_PROGRESS':
                cli_entered = True
                cli_handler = CLIHandler(client_channel, None, handlers)
                cli_handler.loop()


        if run.lifecycle_state in ('FAILED', 'CANCELED'):
            raise Exception(f"Job failed with status: {run.lifecycle_state}")
        return client_channel.read_json('result.json')
        # return {
        #     'state': run.lifecycle_state,
        #     'run_id': run_id,
        #     'succeeded': run.lifecycle_state == 'SUCCEEDED'
        # }



    def kill(self, run_id):
        df_client = get_df_client(self.region)

        r = df_client.delete_run(run_id)
        check_response(r, lambda : SparkETLKillFailure("dataflow failed to kill the run"))



class AbstractJobSubmitter:
    def __init__(self, config):
        self.config = config

    def run(self, deployment_location, options={}, args={}):
        """
        run an application

        Parameters
        -----------
        deployment_location: str
            the location the application build is deployed.
        options: dict
            The runtime options, it could be vendor specific. The job submitter understands it. Not passed to application.
        args: dict
            the runtime argument that passed to application.
        
        return:
        it should return a dict with 2 keys.
            state :  If the job succeeded, it should be SUCCEEDED, otherwise it should be FAILED
            run_id:  A string of unique ID of the run
        """
        raise NotImplementedError()



import os
import json
import time
import subprocess
from urllib.parse import urlparse
import uuid
import tempfile

from requests.auth import HTTPBasicAuth
import requests

from .abstract_job_submitter import AbstractJobSubmitter
from spark_etl.utils import CLIHandler
from spark_etl.core import ClientChannelInterface
from spark_etl.exceptions import SparkETLLaunchFailure

class ClientChannel(ClientChannelInterface):
    def __init__(self, bridge, stage_dir, run_dir, run_id):
        self.bridge = bridge        # the bridge server's name which we can ssh to
        self.stage_dir = stage_dir  # stage_dir is on bridge
        self.run_dir = run_dir      # base dir for all runs, e.g. hdfs:///beta/etl/runs
        self.run_id  = run_id       # string, unique id of the run


    def _create_stage_dir(self):
        # We are using stage_dir instead of temp dir to stage objects on bridge
        # the reason is, many cases temp dir is in memory and cannot hold large object
        subprocess.check_call(
            [
                "ssh", "-q", self.bridge,
                "mkdir", "-p", os.path.join(self.stage_dir, self.run_id)
            ],
            stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
        )


    def read_json(self, name):
        """read a json object from HDFS run dir with specific name
        """
        self._create_stage_dir()

        # copy to stage dir on bridge first
        subprocess.check_call(
            [
                "ssh", "-q", self.bridge,
                "hdfs", "dfs", "-copyToLocal", "-f",
                os.path.join(self.run_dir, self.run_id, name),
                os.path.join(self.stage_dir, self.run_id, name)
            ],
            stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
        )

        try:
            stage_file = tempfile.NamedTemporaryFile(mode="w", delete=False)
            stage_file.close()

            subprocess.check_call(
                [
                    "scp", "-q",
                    os.path.join(
                        f"{self.bridge}:{self.stage_dir}", self.run_id, name
                    ),
                    stage_file.name
                ],
                stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
            )

            with open(stage_file.name) as f:
                return json.load(f)
        finally:
            os.remove(stage_file.name)


    def has_json(self, name):
        """check if a file with specific name exist or not in the HDFS run dir
        """
        exit_code = subprocess.call(
            [
                "ssh", "-q", self.bridge,
                "hdfs", "dfs", "-test", "-f",
                os.path.join(self.run_dir, self.run_id, name)
            ],
            stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
        )
        if exit_code == 0:
            return True
        if exit_code == 1:
            return False
        raise Exception(f"Unrecognized exit code: {exit_code}")


    def write_json(self, name, payload):
        """write a json object to HDFS run dir with specific name
        """
        self._create_stage_dir()

        stage_file = tempfile.NamedTemporaryFile(mode="w", delete=False)
        stage_file.close()
        try:
            with open(stage_file.name, "wt") as f:
                json.dump(payload, f)

            # copy over to bridge at stage directory
            subprocess.check_call(
                [
                    "scp", "-q",
                    stage_file.name,
                    os.path.join(
                        f"{self.bridge}:{self.stage_dir}", self.run_id, name
                    )
                ],
                stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
            )

            # then upload to HDFS, -f for overwrite if exist
            subprocess.check_call(
                [
                    "ssh", "-q", self.bridge,
                    "hdfs", "dfs", "-copyFromLocal", "-f",
                    os.path.join(self.stage_dir, self.run_id, name),
                    os.path.join(self.run_dir, self.run_id, name)
                ],
                stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
            )
        finally:
            os.remove(stage_file.name)


    def delete_json(self, name):
        subprocess.check_call(
            [
                "ssh", "-q", self.bridge,
                "hdfs", "dfs", "-rm",
                os.path.join(self.run_dir, self.run_id, name)
            ],
            stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT
        )


class LivyJobSubmitter(AbstractJobSubmitter):
    def __init__(self, config):
        super(LivyJobSubmitter, self).__init__(config)

    # options is vendor specific arguments
    # args.conf is the spark job config
    # args is arguments need to pass to the main entry
    def run(self, deployment_location, options={}, args={}, handlers=[], on_job_submitted=None, cli_mode=False):
        bridge = self.config["bridge"]
        stage_dir = self.config['stage_dir']
        run_dir = self.config['run_dir']

        # create run dir
        run_id = str(uuid.uuid4())
        subprocess.check_call([
            "ssh", "-q", bridge,
            "hdfs", "dfs", "-mkdir", "-p",
            os.path.join(run_dir, run_id)
        ])

        client_channel = ClientChannel(bridge, stage_dir, run_dir, run_id)
        client_channel.write_json("input.json", args)

        o = urlparse(deployment_location)
        if o.scheme not in ('hdfs', 's3'):
            raise SparkETLLaunchFailure("deployment_location must be in hdfs or s3")

        headers = {
            "Content-Type": "application/json",
            "X-Requested-By": "root",
            'proxyUser': 'root'
        }

        config = {
            'file': os.path.join(deployment_location, "job_loader.py"),
            'pyFiles': [ os.path.join(deployment_location, "app.zip") ],
            'files': [ os.path.join(deployment_location, "lib.zip") ],
            'args': [
                '--run-id', run_id,
                '--run-dir', os.path.join(run_dir, run_id),
            ]
        }
        base_lib_dir = self.config.get('base_lib_dir')
        if base_lib_dir:
            config['args'].extend(['--base-lib-dir', base_lib_dir])

        config.update(options)
        config.pop("display_name", None)  # livy job submitter does not support display_name

        service_url = self.config['service_url']
        username = self.config['username']
        password = self.config['password']

        # print(json.dumps(config))
        r = requests.post(
            os.path.join(service_url, "batches"),
            data=json.dumps(config),
            headers=headers,
            auth=HTTPBasicAuth(username, password),
            verify=False
        )
        if r.status_code not in [200, 201]:
            msg = "Failed to submit the job, status: {}, error message: \"{}\"".format(
                r.status_code,
                r.content
            )
            raise Exception(msg)

        print(f'job submitted, run_id = {run_id}')
        ret = json.loads(r.content.decode("utf8"))
        if on_job_submitted is not None:
            on_job_submitted(run_id, vendor_info=ret)
        job_id = ret['id']
        print('job id: {}'.format(job_id))
        print(ret)
        print('logs:')
        for log in ret.get('log', []):
            print(log)

        def get_job_status():
            r = requests.get(
                os.path.join(service_url, "batches", str(job_id)),
                headers=headers,
                auth=HTTPBasicAuth(username, password),
                verify=False
            )
            if r.status_code != 200:
                msg = "Failed to get job status, status: {}, error message: \"{}\"".format(
                    r.status_code,
                    r.content
                )
                raise Exception(msg)
            ret = json.loads(r.content.decode("utf8"))
            return ret

        cli_entered = False
        # pull the job status
        while True:
            ret = get_job_status()
            job_state = ret['state']
            appId = ret.get('appId')
            print('job_state: {}, applicationId: {}'.format(
                job_state, appId

            ))
            if job_state in ['success', 'dead', 'killed']:
                # cmd = f"yarn logs -applicationId {appId}"
                # host = self.config['bridge']
                # subprocess.call(["ssh", "-q", "-t", host, cmd], shell=False)
                print(f"job_state={job_state}")
                if job_state == 'success':
                    return client_channel.read_json("result.json")
                raise Exception("Job failed")
            time.sleep(5)

            # we enter the cli mode upon first reached the running status
            if cli_mode and not cli_entered and job_state == 'running':
                cli_entered = True
                cli_handler = CLIHandler(client_channel, lambda : get_job_status()['state'] not in ('success', 'dead', 'killed'))
                cli_handler.loop()





import subprocess
import os
import shutil


from spark_etl.deployers import AbstractDeployer
from spark_etl import Build


class LocalDeployer(AbstractDeployer):
    """
    This deployer deploys application to local file system
    """
    def __init__(self, config):
        super(LocalDeployer, self).__init__(config)

    def deploy(self, build_dir, deployment_location):
        # directory layout
        # deployment_location
        #   |
        #   +-- {build_version}
        #         |
        #         +-- app.zip
        #         |
        #         +-- job_loader.py
        #         |
        #         +-- lib
        #         |
        #         +-- main.py
        #         |
        #         +-- manifest.json

        build = Build(build_dir)

        # copy artifacts
        target_dir = os.path.join(deployment_location, build.version)
        if os.path.isdir(target_dir):
            shutil.rmtree(target_dir)
        os.makedirs(target_dir)
        for artifact in build.artifacts:
            src = os.path.join(build_dir, artifact)
            dst = os.path.join(target_dir, artifact)
            print(f"Copy from {src} to {dst}")
            shutil.copyfile(src, dst)

        src = os.path.join(os.path.dirname(os.path.abspath(__file__)), "job_loader.py")
        dst = os.path.join(target_dir, "job_loader.py")
        shutil.copyfile(src, dst)
        print(f"Copy from {src} to {dst}")

        # now unzip the lib
        os.makedirs(os.path.join(target_dir, 'lib'))
        subprocess.check_call([
            'unzip',
            os.path.join(target_dir, 'lib.zip'),
            '-d',
            os.path.join(target_dir, 'lib')
        ])
        os.remove(os.path.join(target_dir, 'lib.zip'))
        print("Deployment is done")







import importlib
import argparse
import json
import os
import subprocess
import sys

from pyspark.sql import SparkSession

def _bootstrap():
    parser = argparse.ArgumentParser(description='job')
    parser.add_argument(
        "--run-id", type=str, required=True, help="Run ID",
    )
    parser.add_argument(
        "--run-dir", type=str, required=True, help="Run Directory",
    )
    parser.add_argument(
        "--app-dir", type=str, required=True, help="Application Directory",
    )
    args = parser.parse_args()
    print(f"run-id:  {args.run_id}")
    print(f"run-dir: {args.run_dir}")
    print(f"app-dir: {args.app_dir}")

    run_home = os.path.join(args.run_dir, args.run_id)
    print(f"run-home: {run_home}")
    os.chdir(run_home)

    # setup lib path
    sys.path.insert(0, os.path.join(args.app_dir, 'app.zip'))
    sys.path.insert(0, os.path.join(args.app_dir, 'lib'))

    # change home dir to the deployment location
    os.chdir(args.app_dir)

    # load input args
    with open(os.path.join(args.run_dir, args.run_id, "input.json"), "r") as f:
        input_args = json.load(f)

    spark = SparkSession.builder.appName(args.run_id).getOrCreate()

    try:
        entry = importlib.import_module("main")
        result = entry.main(spark, input_args)

        # save output
        with open(os.path.join(run_home, "result.json"), "w") as out_f:
            json.dump(result, out_f)

    finally:
        spark.stop()

_bootstrap()




import os
from urllib.parse import urlparse

from .abstract_deployer import AbstractDeployer
from spark_etl import Build
from spark_etl.exceptions import SparkETLDeploymentFailure

import boto3
from botocore.config import Config

class S3Deployer(AbstractDeployer):
    """
    This deployer deploys application to AWS S3 buckets
    """
    def __init__(self, config):
        super(S3Deployer, self).__init__(config)


    def deploy(self, build_dir, deployment_location):
        o = urlparse(deployment_location)
        if o.scheme != 's3':
            raise SparkETLDeploymentFailure("deployment_location must be in s3")

        build = Build(build_dir)

        s3_client = boto3.client(
            's3',
            aws_access_key_id=self.config['aws_access_key_id'],
            aws_secret_access_key=self.config['aws_secret_access_key']
        )
        bucket_name = o.netloc
        s3_dirname = os.path.join(o.path[1:], build.version)

        print(f"Upload to AWS s3, bucket name = {bucket_name}")
        for artifact in build.artifacts:
            local_filename = os.path.join(build.build_dir, artifact)
            object_name = os.path.join(s3_dirname, artifact)

            print(f"{local_filename}  ==> {object_name}")
            s3_client.upload_file(local_filename, bucket_name, object_name)


        local_filename = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'job_loader.py')
        object_name = os.path.join(s3_dirname, "job_loader.py")
        print(f"{local_filename}  ==> {object_name}")
        s3_client.upload_file(local_filename, bucket_name, object_name)

